# Copyright (C) 2006-2007 Søren Roug, European Environment Agency
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
#
# Contributor(s):
#


from .namespaces import ANIMNS
from .element import Element


# Autogenerated
def Animate(**args):
    return Element(qname=(ANIMNS,'animate'), **args)


def Animatecolor(**args):
    return Element(qname=(ANIMNS,'animateColor'), **args)


def Animatemotion(**args):
    return Element(qname=(ANIMNS,'animateMotion'), **args)


def Animatetransform(**args):
    return Element(qname=(ANIMNS,'animateTransform'), **args)


def Audio(**args):
    return Element(qname=(ANIMNS,'audio'), **args)


def Command(**args):
    return Element(qname=(ANIMNS,'command'), **args)


def Iterate(**args):
    return Element(qname=(ANIMNS,'iterate'), **args)


def Par(**args):
    return Element(qname=(ANIMNS,'par'), **args)


def Param(**args):
    return Element(qname=(ANIMNS,'param'), **args)


def Seq(**args):
    return Element(qname=(ANIMNS,'seq'), **args)


def Set(**args):
    return Element(qname=(ANIMNS,'set'), **args)


def Transitionfilter(**args):
    return Element(qname=(ANIMNS,'transitionFilter'), **args)
